import json
import os

import pkg_resources
import requests
from elasticsearch import VERSION as ES_VERSION
from invenio_base.signals import app_loaded
from jsonpointer import JsonPointer, JsonPointerException
from pkg_resources import iter_entry_points

from oarepo_mapping_includes.mapping_transformer import mapping_transformer


class IncludedMapping:
    def __init__(self, name, json):
        self.name = name
        self.json = json
        self.build_ids()

    def get_id(self, idd):
        if not idd:
            return self.json
        try:
            return self.ids[idd]
        except KeyError:
            raise KeyError('Id %s not found in %s' % (idd, self.name))

    def get_pointer(self, pointer):
        if not pointer:
            return self.json

        try:
            return JsonPointer(pointer).resolve(self.json)
        except JsonPointerException as e:
            raise KeyError('Json pointer %s not found in %s: %s' % (pointer, self.name, str(e)))

    def build_ids(self):
        def build(el):
            el_id = el.pop('$id', None)
            if el_id:
                self.ids[el_id] = el
            for k, v in el.items():
                if isinstance(v, dict):
                    build(v)

        self.ids = {}
        build(self.json)


class OARepoMappingIncludesState:
    def __init__(self, app, _includes=None, _handlers=None):
        self.app = app
        self._includes = _includes
        self._handlers = _handlers

    @property
    def included_mappings(self):
        if self._includes is None:
            self._includes = self._load_included_mappings()
        return self._includes

    @property
    def mapping_handlers(self):
        if self._handlers is None:
            self._handlers = self._load_mapping_handlers()
        return self._handlers

    def _load_included_mappings(self):
        included_mappings = {}
        for ep in iter_entry_points('oarepo_mapping_includes'):
            package_name = '{}.v{}'.format(ep.module_name, ES_VERSION[0])
            package_name = package_name.split('.', maxsplit=1)
            package_path = package_name[1].replace('.', '/')
            for filename in pkg_resources.resource_listdir(package_name[0], package_path):
                if filename.endswith('.json'):
                    file_data = pkg_resources.resource_string(
                        package_name[0],
                        os.path.join(package_path, filename))
                    included_mappings[filename] = IncludedMapping(filename, json.loads(file_data.decode("utf-8")))
        return included_mappings

    def _load_mapping_handlers(self):
        handlers = {}
        for ep in iter_entry_points('oarepo_mapping_handlers'):
            handlers[ep.name] = ep.load()
        return handlers

    def load_type(self, type, content, root, content_pointer):
        mapping_handlers = self.mapping_handlers

        if '#' not in type:
            resource, json_pointer = type, None
        else:
            resource, json_pointer = type.split('#', maxsplit=1)

        if json_pointer and json_pointer[0] != '/':
            idd = json_pointer
            json_pointer = None
        else:
            idd = None

        if type in mapping_handlers:
            return mapping_handlers[type](
                type=type, resource=resource, id=idd, json_pointer=json_pointer,
                app=self.app, content=content, root=root, content_pointer=content_pointer)

        if resource in mapping_handlers:
            return mapping_handlers[resource](
                type=type, resource=resource, id=idd, json_pointer=json_pointer,
                app=self.app, content=content, root=root, content_pointer=content_pointer)

        mappings = self.included_mappings
        if resource not in mappings:
            if resource.startswith('http://') or resource.startswith('https://'):
                mappings[resource] = IncludedMapping(resource, requests.get(resource).json())
        try:
            mapping = mappings[resource]
        except KeyError:
            if not json_pointer and not idd:
                return None
            raise KeyError('%s not in %s' % (resource, list(sorted(mappings.keys()))))

        if idd:
            return mapping.get_id(idd)
        else:
            return mapping.get_pointer(json_pointer)


class OARepoMappingIncludesExt:

    def __init__(self, app=None):
        """Extension initialization."""
        if app:
            self.init_app(app)

    def init_app(self, app):
        self.init_config(app)
        app.extensions['oarepo-mapping-includes'] = OARepoMappingIncludesState(app)
        app_loaded.connect(mapping_transformer)

    def init_config(self, app):
        pass



# -*- coding: utf-8 -*-
"""Setup module for elasticsearch mapping includes."""
import os

from setuptools import setup

readme = open('README.md').read()

install_requires = [
    'elasticsearch',
    'deepmerge',
    'NamedAtomicLock'
]

tests_require = [
    'requests-mock'
]

extras_require = {
    'tests': [
        *tests_require,
        'oarepo[tests]',
    ]
}

g = {}
with open(os.path.join('oarepo_mapping_includes', 'version.py'), 'rt') as fp:
    exec(fp.read(), g)
    version = g['__version__']

setup(
    name="oarepo-mapping-includes",
    version=version,
    url="https://github.com/oarepo/oarepo-mapping-includes",
    license="MIT",
    author="Miroslav Simek",
    author_email="miroslav.simek@vscht.cz",
    description="An inclusion mechanism for elasticsearch mappings",
    long_description=readme,
    long_description_content_type='text/markdown',
    zip_safe=False,
    packages=['oarepo_mapping_includes'],
    entry_points={
        'invenio_base.apps': [
            'oarepo_mapping_includes = oarepo_mapping_includes.ext:OARepoMappingIncludesExt'
        ],
        'invenio_base.api_apps': [
            'oarepo_mapping_includes = oarepo_mapping_includes.ext:OARepoMappingIncludesExt'
        ]
    },
    include_package_data=True,
    setup_requires=install_requires,
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require=extras_require,
    platforms='any',
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Development Status :: 4 - Beta',
    ],
)



import copy
import json
import os

from NamedAtomicLock import NamedAtomicLock
from deepmerge import conservative_merger

from oarepo_mapping_includes import Mapping

inherited_merger = conservative_merger


def process_type(prop, field, includes, add_field=True, root=None, content_pointer=None, processed_types=None):
    # get the type - it can be a simple type or an array (think of multiple inheritance)
    mapping_types = prop.pop(field, None)
    if not mapping_types:
        return prop, False  # strange, should never happen

    if not isinstance(mapping_types, list):
        mapping_types = [mapping_types]

    # will contain merged parent types
    modified = False
    if processed_types is None:
        processed_types = set()

    orig_mapping_types = [*mapping_types]

    while mapping_types:
        mpt = mapping_types.pop(0)
        # for each of the mapping type
        if mpt in processed_types:
            continue

        processed_types.add(mpt)

        # try to load the type
        mapping = includes.load_type(mpt, content=prop, root=root, content_pointer=content_pointer)

        if not mapping:
            continue
        modified = True

        # if it is an instance of loaded mapping and it already took care of extra data in original,
        # just replace the original
        if isinstance(mapping, Mapping) and not mapping.merge:
            prop.clear()
            prop.update(copy.deepcopy(mapping.mapping))
        else:
            # otherwise merge with the original
            if isinstance(mapping, Mapping):
                mapping = mapping.mapping

            # merge into mpt_res, overwriting any previously existing values
            inherited_merger.merge(prop, mapping)

        # oarepo:type is a special construct to use the type but break recursion
        new_type = prop.pop('oarepo:type', None)
        if new_type:
            prop[field] = new_type
            continue

        # extract the new type
        new_type = prop.get(field, None)
        if not new_type:
            continue

        mpt = new_type
        if not isinstance(mpt, (tuple, list)):
            mpt = [mpt]
        mapping_types = mpt + mapping_types

    # if included mappings do not add type, use the first one in the original type
    if add_field and field not in prop and not modified:
        prop[field] = orig_mapping_types[0]

    return prop, modified  # prop is for tests


def convert_props(includes, properties, root, content_pointer):
    # for each property, update its type and definition if needed
    if 'oarepo:extends' in properties:
        for i in range(100):
            modified = process_type(properties, 'oarepo:extends', includes, False,
                         root=root, content_pointer=content_pointer)[1]
            if not modified:
                break

    for name, prop in properties.items():
        prop_pointer = content_pointer + '/' + name
        if isinstance(prop, str):
            continue

        for i in range(100):
            modified = process_type(prop, 'oarepo:extends', includes, False, root=root, content_pointer=prop_pointer)[1]
            modified = process_type(prop, 'type', includes, True, root, prop_pointer)[1] or modified
            if not modified:
                break
        else:
            raise Exception('Infinite recursion in mapping type at path %s' % content_pointer)

        # and go recursively
        convert_props(includes, prop.get('properties', {}), root, prop_pointer + '/properties')
        convert_props(includes, prop.get('fields', {}), root, prop_pointer + '/fields')


def convert_extends(includes, el, root, content_pointer):
    for name, prop in el.items():
        if not isinstance(prop, dict):
            continue
        prop_pointer = content_pointer + '/' + name

        process_type(prop, 'oarepo:extends', includes, False, root=root, content_pointer=prop_pointer)
        prop.pop('oarepo:extends', None)
        # no need to go recursively as it is called in convert_props as well
        # convert_extends(includes, prop, root, prop_pointer)


def process(includes, base_dir, filename):
    # load the file, convert its types and write it back into cache directory
    dest_file = os.path.join(base_dir, os.path.basename(filename))
    with open(filename) as f:
        mapping = json.load(f)
    convert_extends(includes, mapping, root=mapping, content_pointer='')
    convert_props(includes, mapping['mappings']['properties'], mapping, '/mappings/properties')
    mapping_to_save = json.dumps(mapping, ensure_ascii=False, indent=4)
    if os.path.exists(dest_file):
        with open(dest_file, 'r') as f:
            existing = f.read()
            if existing == mapping_to_save:
                # do not save again the same data
                return dest_file
    with open(dest_file, 'w') as f:
        f.write(mapping_to_save)

    return dest_file


def mapping_transformer(source, app=None, **kwargs):
    includes = app.extensions['oarepo-mapping-includes']
    mappings = app.extensions['invenio-search'].mappings

    lock = NamedAtomicLock('oarepo-mapping-includes')
    lock.acquire()
    try:
        # create cache directory
        transformed_mappings_dir = os.path.join(app.instance_path, 'mappings')
        if not os.path.exists(transformed_mappings_dir):
            os.makedirs(transformed_mappings_dir)

        # and transform each mapping
        for k, v in list(mappings.items()):
            mappings[k] = process(includes, transformed_mappings_dir, v)
    finally:
        lock.release()
